"""
Operational semantics for the Why language

Provides evaluation functions for the three syntactic classes,
based on a environment mapping variables to integer values.
These functions return None when an error occur somewhere (Big step semantics)
"""

from syntax import *

def eval_aexp(env: dict[str,int], exp: ArithExpr) -> int | None:
    """evaluates an arithmetic expression"""
    match exp:
        case AECst(value):
            return value
        case AEVar(var):
            if(var in env): return env[var]
            else: return 0
        case AEUop(uop,expr):
            match uop:
                case Uop.OPP:
                    return -1*eval_aexp(env, expr)
                case _:
                    assert False
        case AEBop(bop,left_expr,right_expr):
            match bop:
                case Bop.ADD:
                    return eval_aexp(env, left_expr)+eval_aexp(env, right_expr)
                case Bop.MUL:
                    return eval_aexp(env, left_expr)*eval_aexp(env, right_expr)
                case Bop.DIV:
                    if (eval_aexp(env, right_expr)!=0): return eval_aexp(env, left_expr)/eval_aexp(env, right_expr) 
                    else: return None
                case _:
                    assert False
        case _: assert False

def eval_bexp(env: dict[str,int], exp:BoolExpr) -> bool | None:
    """evaluates a boolean expression"""
    match exp:
        case BEPlain(aexpr):
            return eval_aexp(env, aexpr)!=0
        case BEEq(left_expr,right_expr):
            return eval_aexp(env, left_expr) == eval_aexp(env, right_expr)
        case BELeq(left_expr,right_expr):
            return eval_aexp(env, left_expr) <= eval_aexp(env, right_expr)
        case BENeg(expr):
            return not(eval_bexp(env, expr))
        case _: assert False

def eval_instr(env: dict[str,int], instr: Instr) -> dict[str,int] | None:
    """evaluates an instruction"""
    match instr:
        case ISkip():
            return env
        case ISet(var,expr):
            newenv = env.copy()
            newenv[var]=eval_aexp(env, expr); return newenv
        case ISeq(first,second):
            return eval_instr(eval_instr(env, first), second)
        case ICond(cond,true_branch,false_branch):
            if eval_bexp(env, cond): return eval_instr(env, true_branch)
            else: return eval_instr(env, false_branch)
        case ILoop(cond,body):
            if eval_bexp(env, cond): return eval_instr(eval_instr(env, body), ILoop(cond,body))
            else: return env
        case _: assert False
